/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: icpciIoctl.c,v 5.1.2.6 1999/09/22 11:47:13 matthias Exp $
 *
 */

/*
 * Defines:
 */

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */

/*
 * Globals:
 */

extern struct 
icp_global_t icp;	/* main driver structure */

int rcsrevision = 0x00050100;	/* means 0.5.1 */

/*
 * Prototypes:
 */

/*
 * icp_ioctl
 */

int
icp_ioctl (struct icp_devno_t *devno,
	   icp_flags_t flags, icp_f_owner_t *f_owner, u_int cmd,
	   union icp_inparam_t *inparam, union icp_outparam_t *outparam)
{

  int error, count;

  u_char board, module, camera;

  void *src_addr, *dst_addr;
  off_t	src_off, dst_off;

  long extime;
  u_char exclk;
  

#ifdef DEBUG_ICPCI
  const struct ioctl_cmd_t 
  {
    u_int number;
    char text[32];
  } 
  ioctl_cmd[] =
  {
    { GIOC_CR_SET_FREEZE,	"CR_SET_FREEZE" },
    { GIOC_CR_SET_GRAB,		"CR_SET_GRAB" },
    { GIOC_CR_SET_SNAP,		"CR_SET_SNAP" },
    { GIOC_CR_SET_NORMAL,	"CR_SET_NORMAL" },
    { GIOC_CR_SET_TRIGGER,	"CR_SET_TRIGGER" },
    { GIOC_CR_SET_SOFTTRIG,	"CR_SET_SOFTTRIG" },
    { GIOC_CR_SET_HVDEC,	"CR_SET_HVDEC" },
    { GIOC_VS_GET_LUT, 		"VS_GET_LUT" },
    { GIOC_VS_SET_LUT, 		"VS_SET_LUT" },
    { GIOC_VS_SET_LUT_LIN,	"VS_SET_LUT_LIN" },
    { GIOC_VS_GET_NREF, 	"VS_GET_NREF" },
    { GIOC_VS_SET_NREF, 	"VS_SET_NREF" },
    { GIOC_VS_GET_PREF, 	"VS_GET_PREF" },
    { GIOC_VS_SET_PREF, 	"VS_SET_PREF" },
    { GIOC_DG_GET_LUTPG,	"DG_GET_LUTPG" },
    { GIOC_DG_SET_LUTPG,	"DG_SET_LUTPG" },       
    { GIOC_DG_GET_LUT,		"DG_GET_LUT" },         
    { GIOC_DG_SET_LUT,		"DG_SET_LUT" },         
    { GIOC_DG_GET_OLUTPG,	"DG_GET_OLUTPG" },      
    { GIOC_DG_SET_OLUTPG,	"DG_SET_OLUTPG" },      
    { GIOC_DG_GET_OLUT,		"DG_GET_OLUT" },        
    { GIOC_DG_SET_OLUT,		"DG_SET_OLUT" },        
    { GIOC_DG_SET_LUT_LIN,	"DG_SET_LUT_LIN" },        
    { GIOC_DG_SET_NORM_TO8,	"DG_SET_NORM_TO8" },
    { GIOC_DG_SET_NORM_TO16,	"DG_SET_NORM_TO16" },        
    { GIOC_DG_GET_EXTIME,	"DG_GET_EXTIME" },
    { GIOC_DG_SET_EXTIME,	"DG_SET_EXTIME" },        
    { GIOC_GE_GET_MODULE,	"GE_GET_MODULE" },
    { GIOC_GE_GET_VERSION,	"GE_GET_VERSION" },
    { GIOC_GE_GET_BMAMAPSIZE, 	"GE_GET_BMAMAPSIZE" },
    { GIOC_GE_GET_BMABUFFERS, 	"GE_GET_BMABUFFERS" },
    { GIOC_GE_GET_BMAWIDTH, 	"GE_GET_BMAWIDTH" },
    { GIOC_GE_GET_BMAHEIGHT, 	"GE_GET_BMAHEIGHT" },
    { GIOC_GE_GET_BMABUFSIZE, 	"GE_GET_BMABUFSIZE" },
    { GIOC_GE_GET_BMABUF, 	"GE_GET_BMABUF" },
    { GIOC_GE_SET_BMABUF, 	"GE_SET_BMABUF" },
    { GIOC_GE_GET_BMABUFADDR, 	"GE_GET_BMABUFADDR" },
    { GIOC_GE_GET_MEMMAPSIZE, 	"GE_GET_MEMMAPSIZE" },
    { GIOC_GE_GET_MEMBUFFERS, 	"GE_GET_MEMBUFFERS" },
    { GIOC_GE_GET_MEMWIDTH, 	"GE_GET_MEMWIDTH" },
    { GIOC_GE_GET_MEMHEIGHT, 	"GE_GET_MEMHEIGHT" },
    { GIOC_GE_GET_MEMBUFSIZE, 	"GE_GET_MEMBUFSIZE" },
    { GIOC_GE_GET_MEMBUF, 	"GE_GET_MEMBUF" },
    { GIOC_GE_SET_MEMBUF, 	"GE_SET_MEMBUF" },
    { GIOC_GE_GET_MEMBUFADDR, 	"GE_GET_MEMBUFADDR" },
    { GIOC_GE_GET_CAMERA, 	"GE_GET_CAMERA" },
    { GIOC_GE_SET_CAMERA, 	"GE_SET_CAMERA" },
    { GIOC_GE_GET_CAMCNF, 	"GE_GET_CAMCNF" },
    { GIOC_GE_SET_CAMCNF, 	"GE_SET_CAMCNF" },
    { GIOC_GE_GET_FPS,		"GE_GET_FPS" },
    { GIOC_GE_SET_FPS,		"GE_SET_FPS" },
    { GIOC_GE_GET_STATS,	"GE_GET_STATS" },
    { GIOC_GE_SET_STATS,	"GE_SET_STATS" },
    { HIOC_PV_GET_LUT, 		"PV_GET_LUT" },
    { HIOC_PV_SET_LUT, 		"PV_SET_LUT" },
    { HIOC_PV_SET_LUT_LIN,	"PV_SET_LUT_LIN" },
    { HIOC_PV_GET_NREF, 	"PV_GET_NREF" },
    { HIOC_PV_SET_NREF, 	"PV_SET_NREF" },
    { HIOC_PV_GET_PREF, 	"PV_GET_PREF" },
    { HIOC_PV_SET_PREF, 	"PV_SET_PREF" },
    { 0, " " }
  };

  char cmd_text[32];
  int  index = 0;

  strcpy (cmd_text, "COMMAND NOT SUPPORTED");
  while (ioctl_cmd[index].number != 0)
    {
      if (ioctl_cmd[index].number == cmd)
	{
	  strcpy (cmd_text, ioctl_cmd[index].text);
	  break;
	}
      index++;
    }
  ICP_PRINTF("icpci[%c]: icp_ioctl command: %s.\n", 'F', cmd_text);
#endif

  board = devno->board;
  module = devno->module;
  camera = devno->camera;

  /* argument validation */
  if (camera != ICP_CAM_IOCTL)
    {
      /* NOT ALLOWED on normal devices */
      switch (cmd)
	{
	case GIOC_GE_SET_BMABUF:
	case GIOC_GE_SET_MEMBUF:
	case GIOC_GE_SET_CAMERA:
	case GIOC_CR_SET_FREEZE:
	case GIOC_CR_SET_GRAB:
	case GIOC_CR_SET_SNAP:
	case GIOC_CR_SET_NORMAL:
	case GIOC_CR_SET_TRIGGER:
	case GIOC_CR_SET_SOFTTRIG:
	  return -EINVAL;
	}
    }

  if (module != ICP_MOD_IDENT)
    {
      /* no further information on module predefined devices */
      switch (cmd)
	{
	case GIOC_GE_GET_MODULE:
	  return -EINVAL;
	}
    }

  if (!icp.bma_on)
    {
      /* functions only available if bma on */
      switch (cmd)
	{
	case GIOC_GE_GET_BMAMAPSIZE:
	case GIOC_GE_GET_BMABUFFERS:
	case GIOC_GE_GET_BMAWIDTH:
	case GIOC_GE_GET_BMAHEIGHT:
	case GIOC_GE_GET_BMABUFSIZE:
	case GIOC_GE_GET_BMABUF:
	case GIOC_GE_SET_BMABUF:
	case GIOC_GE_GET_BMABUFADDR:
	  return -EINVAL;
	}
    }

  switch (cmd)
    {

      /* GENERIC */
      
      /* get software version */
    case GIOC_GE_GET_VERSION:
      outparam->get_version = rcsrevision;
      break;

      /* get module type */
    case GIOC_GE_GET_MODULE:
      outparam->get_module = icp.cap[board]->module;
      break;

      /* get width of bma-buffer */
    case GIOC_GE_GET_BMAWIDTH:
      outparam->get_bmawidth = icp.bma[board]->width;
      break;

      /* get height of bma-buffer */
    case GIOC_GE_GET_BMAHEIGHT:
      outparam->get_bmaheight = icp.bma[board]->height;
      break;

      /* get size of upper memory area */
    case GIOC_GE_GET_BMAMAPSIZE:
      outparam->get_bmamapsize = icp.bma[board]->buf_size;
      break;

      /* get number of bus master target buffers */
    case GIOC_GE_GET_BMABUFFERS:
      outparam->get_bmabuffers = icp.bma[board]->buffers;
      break;

      /* get size (page aligned with struct icp_info_t) of every bma-buffer */
    case GIOC_GE_GET_BMABUFSIZE:
      outparam->get_bmabufsize = icp.bma[board]->paged_size;
      break;

      /* get current bma-buffer */
    case GIOC_GE_GET_BMABUF:
      outparam->get_bmabuf = icp.bma[board]->cur_buf;
      break;

      /* set current bma-buffer */ 
    case GIOC_GE_SET_BMABUF:
      ICP_PRINTF("icpci[%d]:       bma-buffer : %d.\n",
		 board, inparam->set_bmabuf);
      if (inparam->set_bmabuf >= icp.bma[board]->buffers)
	{
	  error = -EINVAL;
	  goto error;
	}
      icp_bma_clear (board, ICP_CLEAR_BLACK);
      icp.bma[board]->cur_buf = inparam->set_bmabuf;
      break;

      /* get offset address of active bma-buffer */
    case GIOC_GE_GET_BMABUFADDR:
      outparam->get_bmabufaddr =
	(u_long)icp.bma[board]->buffer[icp.bma[board]->cur_buf];
      ICP_PRINTF("icpci[%d]:       bma-address: %#06lx.\n",
		 board, outparam->get_bmabufaddr);
      break;

      /* get size of on board memory area */
    case GIOC_GE_GET_MEMMAPSIZE:
      outparam->get_memmapsize = (u_long)icp.mem[board]->buf_size;
      break;

      /* get number of on board memory buffers */
    case GIOC_GE_GET_MEMBUFFERS:
      outparam->get_membuffers = icp.mem[board]->buffers;
      break;

      /* get width of mem-buffer */
    case GIOC_GE_GET_MEMWIDTH:
      outparam->get_memwidth = icp.mem[board]->width;
      break;

      /* get height of mem-buffer */
    case GIOC_GE_GET_MEMHEIGHT:
      outparam->get_memheight = icp.mem[board]->height;
      break;

      /* get size (without struct icp_info_t) of every mem-buffer */
    case GIOC_GE_GET_MEMBUFSIZE:
      outparam->get_membufsize = icp.mem[board]->raw_size;
      break;

      /* get current mem-buffer */ 
    case GIOC_GE_GET_MEMBUF:
      outparam->get_membuf = icp.mem[board]->cur_buf;
      break;

      /* set current mem-buffer */ 
    case GIOC_GE_SET_MEMBUF:
      ICP_PRINTF("icpci[%d]:       mem-buffer : %d.\n", board,
		 inparam->set_membuf);
      if (inparam->set_membuf >= icp.mem[board]->buffers)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (cr_set_aqstart (board, icp.mem[board]->buffer[inparam->set_membuf]))
	{
	  icp_error (board, "cr_set_aqstart");
	  error = -EIO;
	  goto error;
	}
      icp.mem[board]->cur_buf = inparam->set_membuf;
      icp_mem_clear (board, ICP_CLEAR_BLACK);
      break;

      /* get offset address of active mem-buffer */
    case GIOC_GE_GET_MEMBUFADDR:
      outparam->get_membufaddr =
	(u_long)icp.mem[board]->buffer[icp.mem[board]->cur_buf];
      ICP_PRINTF("icpci[%d]:       mem-address: %#06lx.\n", board,
		 outparam->get_membufaddr);
      break;

      /* get current camera */
    case GIOC_GE_GET_CAMERA:
      outparam->get_camera = icp.cap[board]->cur_cam;
      break;

      /* set current camera */ 
    case GIOC_GE_SET_CAMERA:
      ICP_PRINTF("icpci[%d]:           camera : %d.\n", board,
		 inparam->set_camera);
      if (inparam->set_camera > icp.cap[board]->cameras)
	{
	  error = -EINVAL;
	  goto error;
	}
      /* test for snap cycle in progress */
      if (!icp.complete[board])
	{
	  error = -EBUSY;
	  goto error;
	}
      /* no switching neccessary */
      if (icp.cap[board]->module == ICP_AMDG)
	break;
      /* enable interrupts */
      if ((error = icp_enable_irq (board, CR_VB_FLD)))
	goto error;
      /* disable grabbing/snapping */
      if (cr_get_acqmd (board) != CR_FREEZE)
	{
	  /* set freeze command */
	  if (cr_set_acqmd (board, CR_FREEZE))
	    {
	      icp_error (board, "cr_set_acqmd");
	      error = -EIO;
	      goto error;
	    }
	  /* synchronize to VB */
	  icp.opmode[board] = ICP_OP_VB_FREEZ;
	  if ((error = icp_sleep (board, icp.expire[board], CAMERA)))
	    goto error;
	  icp.opmode[board] = ICP_OP_NO;
	}
      /* clear current mem and bma buffers */
      icp_mem_clear (board, ICP_CLEAR_BLACK);
      if (icp.bma_on)
	icp_bma_clear (board, ICP_CLEAR_BLACK);
      /* switch input port */
      switch (icp.cap[board]->module)
	{
	case ICP_AMVS:
	  if (vs_set_camsel (board, inparam->set_camera))
	    { 
	      icp_error (board, "vs_set_camsel");
	      error = -EIO;
	      goto error;
	    }
	  break;
	case ICP_AMDG:
	  /* no switching, only cam 0 available */
	  break;
	case ICP_AMPV:
	  if (pv_set_camsel (board, inparam->set_camera))
	    { 
	      icp_error (board, "pv_set_camsel");
	      error = -EIO;
	      goto error;
	    }
	  break;
	default:
	  error = -ENXIO;
	  goto error;
	}
      icp.cap[board]->cur_cam = inparam->set_camera;
      /* synchronize to VB */
      icp.opmode[board] = ICP_OP_VB_FREEZ;
      if ((error = icp_sleep (board, icp.expire[board], CAMERA)))
	goto error;
      /* wait for stabilization of signal (or not) */
      icp.opmode[board] = ICP_OP_OTHER;
      if (ICP_WAIT_CAMERA)
	if ((error = icp_sleep (board, icp.expire[board] *
				ICP_WAIT_CAMERA, CAMERA)))
	  goto error;
      icp.opmode[board] = ICP_OP_NO;
      break;

      /* get camera structure */
    case GIOC_GE_GET_CAMCNF:
      switch (icp.cap[board]->module)
	{
	case ICP_AMVS:
	  if (camera == ICP_CAM_IOCTL)
	    /* in this operation mode always use camera 0 */
	    memcpy (&outparam->get_icp_camcnf,
		    &icp.dev[board][0]->vs_cam,
		    sizeof(struct vs_cam_t));
	  else
	    memcpy (&outparam->get_icp_camcnf.vs_cam,
		    &icp.dev[board][camera]->vs_cam,
		    sizeof(struct vs_cam_t));
	  break;
	case ICP_AMDG:
	  memcpy (&outparam->get_icp_camcnf.dg_cam,
		  &icp.dev[board][0]->dg_cam,
		  sizeof(struct dg_cam_t));
	  break;
	case ICP_AMPV:
	  if (camera == ICP_CAM_IOCTL)
	    /* in this operation mode always use camera 0 */
	    memcpy (&outparam->get_icp_camcnf,
		    &icp.dev[board][0]->pv_cam,
		    sizeof(struct pv_cam_t));
	  else
	    memcpy (&outparam->get_icp_camcnf.pv_cam,
		    &icp.dev[board][camera]->pv_cam,
		    sizeof(struct pv_cam_t));
	  break;
	default:
	  error = -ENXIO;
	  goto error;
      }
      break;
      
      /* set camera structure */
    case GIOC_GE_SET_CAMCNF:
      switch (icp.cap[board]->module)
	{
	case ICP_AMVS:
	  if (camera == ICP_CAM_IOCTL)
	    {
	      /* in this operation mode always use camera 0 */
	      memcpy (&icp.dev[board][0]->vs_cam,
		      &inparam->set_icp_camcnf.vs_cam,
		      sizeof(struct vs_cam_t));
	      if (icp_amvs_init (board, 0))
		{
		  error = -EIO;
		  goto error;
		}
	    }
	  else
	    {
	      memcpy (&icp.dev[board][camera]->vs_cam,
		      &inparam->set_icp_camcnf.vs_cam,
		      sizeof(struct vs_cam_t));
	      if (icp_amvs_init (board, camera))
		{
	      error = -EIO;
	      goto error;
		}
	    }
	  break;
	case ICP_AMDG:
	  memcpy (&icp.dev[board][0]->dg_cam,
		  &inparam->set_icp_camcnf.dg_cam,
		  sizeof(struct dg_cam_t));
	  if (icp_amdg_init (board, 0))
	    {
	      error = -EIO;
	      goto error;
	    }
	  switch (dg_get_psize (board))
	    {
	    case DG_PSIZE_8:
	      if (cr_set_pixsz (board, CR_PIXSZ_8))
		{
		  icp_error (board, "cr_set_pixsz");
		  error = -EIO;
		  goto error;
		}
	      break;
	  case DG_PSIZE_10:
	  case DG_PSIZE_12:
	    if (cr_set_pixsz (board, CR_PIXSZ_16))
	      {
		icp_error (board, "cr_set_pixsz");
		error = -EIO;
		goto error;
	      }
	  case DG_PSIZE_16:
	    if (dg_get_ilpixel (board) == DG_ILEAVE)
	      {
		if (cr_set_pixsz (board, CR_PIXSZ_8))
		  {
		    icp_error (board, "cr_set_pixsz");
		    error = -EIO;
		    goto error;
		  }
	      }
	    else
	      {
		if (cr_set_pixsz (board, CR_PIXSZ_16))
		  {
		    icp_error (board, "cr_set_pixsz");
		    error = -EIO;
		    goto error;
		  }
	      }
	    break;
	    default:
	      error = -EIO;
	      goto error;
	    }
	  break;
	case ICP_AMPV:
	  if (camera == ICP_CAM_IOCTL)
	    {
	      /* in this operation mode always use camera 0 */
	      memcpy (&icp.dev[board][0]->pv_cam,
		      &inparam->set_icp_camcnf.pv_cam,
		      sizeof(struct pv_cam_t));
	      if (icp_ampv_init (board, 0))
		{
		  error = -EIO;
		  goto error;
		}
	    }
	  else
	    {
	      memcpy (&icp.dev[board][camera]->pv_cam,
		      &inparam->set_icp_camcnf.pv_cam,
		      sizeof(struct pv_cam_t));
	      if (icp_ampv_init (board, camera))
		{
		  error = -EIO;
		  goto error;
		}
	    }
	  break;
	default:
	  error = -ENXIO;
	  goto error;
	}
      if ((error = icp_mem_init (board)))
	goto error;
      if (icp.bma_on)
	if ((error = icp_bma_init (board)))
	  goto error;
      break;
      
      /* get frames per second value (timeout) */
    case GIOC_GE_GET_FPS:
      outparam->get_fps = (float)HZ / ((float)icp.expire[board] / 1.25);
      break;
      
      /* set frames per second value (timeout) */
    case GIOC_GE_SET_FPS:
      if (inparam->set_fps > ICP_FPS_MAX || inparam->set_fps < ICP_FPS_MIN)
	{
	  error = -EINVAL;
	  goto error;
	}
      icp.expire[board] = (int)(((float)HZ / inparam->set_fps) * 1.25); 
      ICP_PRINTF("icpci[%d]:           expires: %ld.\n", board,
		 icp.expire[board]);
      break;

      /* get some statistics */
    case GIOC_GE_GET_STATS:
      memcpy (&outparam->get_stats, &icp.acc[board],
	      sizeof (struct icp_acc_t));
      break;

      /* reset statistics */
    case GIOC_GE_SET_STATS:
      icp.acc[board].events_captured = 0UL;
      icp.acc[board].bytes_valid = 0UL;
      icp.acc[board].bytes_lost = 0UL;
      break;

    /* IC-PCI */

    /* stop grabbing / snapping */
    case GIOC_CR_SET_FREEZE:
      /* enable interrupts */
      if ((error = icp_enable_irq (board, CR_VB_FLD)))
	goto error;
      if (cr_set_acqmd (board, CR_FREEZE))
	{
	  icp_error (board, "cr_set_acqmd");
	  error = -EIO;
	  goto error;
	}
      /* synchronize to VB */
      icp.opmode[board] = ICP_OP_VB_FREEZ;
      if ((error = icp_sleep (board, icp.expire[board], CAMERA)))
	goto error;
      icp.opmode[board] = ICP_OP_NO;

      /* disable signal assignment */
      icp_deassign_signal (board);

      /* clear task assignment */
      icp_deassign_task (board);

      /* disable used interrupts */
      if ((error = icp_disable_irq (board,
				    CR_VB_FLD | CR_TRIGINT | CR_AQINT)))
	goto error;
      break;

    /* start grabbing */
    case GIOC_CR_SET_GRAB:
      /* test for snap cycle in progress */
      if (!icp.complete[board])
	{
	  error = -EBUSY;
	  goto error;
	}
      /* reset statistics */
      icp_irqoff (board);
      icp.acc[board].frames_captured = 0UL;
      icp.acc[board].frames_valid = 0UL;    
      icp.acc[board].frames_lost = 0UL;    
      icp_irqon (board);
      icp.timeout[board] = FALSE;
      /* assign task (pid) interested in SIGIO */
      icp_assign_signal (board, flags, f_owner);
      /* assign task to check the blocking of SIGIO */
      icp_assign_task (board);

      /* enable interrupts */
      if ((error = icp_enable_irq (board, CR_VB_FLD)))
	goto error;
      /* set grab command */
      if (cr_set_acqmd (board, CR_GRAB))
	{
	  icp_error (board, "cr_set_acqmd");
	  error = -EIO;
	  goto error;
	}
      icp.opmode[board] = ICP_OP_VB_GRAB;
#if (1)
      /* wait for camera signal */
      if ((error = icp_sleep (board, icp.expire[board] * 2, CAMERA)))
	goto error;
      if (icp.timeout[board]) {
	icp.timeout[board] = FALSE;
	icp.opmode[board] = ICP_OP_NO;
	error = -ENODATA;
	goto error;
      }
#endif
      icp.opmode[board] = ICP_OP_GRAB;
      break;

      /* start snapping */
    case GIOC_CR_SET_SNAP:
      ICP_PRINTF("icpci[%d]:           count : %d.\n", board,
		 inparam->set_snap);
      /* argument validation */
      if (inparam->set_snap == 0)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (inparam->set_snap > 0)
	{
	  if (inparam->set_snap >
	      (icp.mem[board]->buffers - icp.mem[board]->cur_buf) ||
	      inparam->set_snap > ICP_SNAPS_MAX)
	    {
	      error = -EINVAL;
	      goto error;
	    }
	} 
      else
	{
	  if (cr_get_tmode (board) == CR_DISABLE)
	    {
	      error = -EINVAL;
	      goto error;
	    }
	  if (-inparam->set_snap >
	      (icp.mem[board]->buffers - icp.mem[board]->cur_buf) ||
	      -inparam->set_snap > ICP_SNAPS_MAX )
	    {
	      error = -EINVAL;
	      goto error;
	    }
	}
      /* test for snap cycle in progress */
      if (!icp.complete[board])
	{
	  error = -EBUSY;
	  goto error;
	}
      icp.timeout[board] = FALSE;
      /* assign task (pid) interested in SIGIO */
      icp_assign_signal (board, flags, f_owner);
      
      /* save current buffer status */
      icp.mem[board]->frst_buf = icp.mem[board]->cur_buf;
      if (icp.bma_on)
	icp.bma[board]->frst_buf = icp.bma[board]->cur_buf;
      /* first, test for signal (if not in Trigger mode) */
      if (cr_get_tmode (board) == CR_DISABLE &&
	  icp.cap[board]->module != ICP_AMDG)
	{
	  icp.opmode[board] = ICP_OP_VB_SNAP;
	  if ((error = icp_enable_irq (board, CR_VBINT)))
	    goto error;
	  if ((error = icp_sleep (board, icp.expire[board], CAMERA)))
	    goto error;
	  if (icp.timeout[board])
	    {
	      icp.timeout[board] = FALSE;
	      icp.opmode[board] = ICP_OP_NO;
	      error = -ENODATA;
	      goto error;
	    }
	  icp.opmode[board] = ICP_OP_NO;
	}
      /* assign the right field first (if not in Trigger mode) */
      /* this is for speed improvement */
      if (cr_get_tmode (board) == CR_DISABLE)
	{
	  if (cr_get_vbstat (board) == CR_OUT_VB)
	    {
	      icp_splhigh (board);
	      if (cr_get_smstat (board) == CR_ILACE)
		{
		  if (cr_get_fldstat (board) == CR_IN_ODD_FIELD)
		    {
		      ICP_PRINTF("icpci[%d]: start snap on even field.\n",
				 board);
		      if (cr_set_fldsel (board, CR_EVEN_FIELD))
			{
			  icp_error (board, "cr_set_fldsel");
			  error = -EIO;
			  goto error;
			}
		    }
		  else
		    {
		      ICP_PRINTF("icpci[%d]: start snap on odd field.\n",
				 board);
		      if (cr_set_fldsel (board, CR_ODD_FIELD))
			{
			  icp_error (board, "cr_set_fldsel");
			  error = -EIO;
			  goto error;
			}
		    }
		}
	      else
		{
		  ICP_PRINTF("icpci[%d]: start snap on next field.\n", board);
		  if (cr_set_fldsel (board, CR_NEXT_FIELD))
		    {
		      icp_error (board, "cr_set_fldsel");
		      error = -EIO;
		      goto error;
		    }
		}
	      icp_spllow (board);
	    }
	  else
	    {
#ifdef DEBUG_ICPCI
	      icp_printf ("icpci[%c]: sample current field in VB?\n", 'W');
#endif
	    }
	}
      error = OK;
      /* reenable right irq */
      if (cr_get_tmode (board) == CR_DISABLE)
	{
	  if ((error = icp_enable_irq (board, CR_VB_FLD | CR_AQINT)))
	    goto error;
	  /* set frame count to requested size */
	  if (cr_set_fcnt (board, inparam->set_snap))
	    {
	      icp_error (board, "cr_set_fcnt");
	      error = -EIO;
	      goto error;
	    }
	}
      else
	{
	  if (inparam->set_snap > 0)
	    {
	      if ((error = icp_enable_irq (board, CR_VBINT)))
		goto error;
	    }
	  else
	    {
	      if ((error = icp_enable_irq (board, CR_VB_FLD |
					   CR_TRIGINT | CR_AQINT)))
		goto error;
	      /* set frame count to requested size */
	      if (cr_set_fcnt (board, -inparam->set_snap))
		{
		  icp_error (board, "cr_set_fcnt");
		  error = -EIO;
		  goto error;
		}
	    }
	}
      /* write magic sequence */
      for (count = 0; count < inparam->set_snap; count++)
	{
	  src_addr = icp.mem[board]->v_buf_base;
	  src_off = icp.mem[board]->buffer[icp.mem[board]->cur_buf + count];
	  icp_writel (ICP_MEM_TEST, ((u_long*)(src_addr + src_off)));;
	  dst_addr = icp.bma[board]->v_buf_base;
	  dst_off = icp.bma[board]->buffer[icp.bma[board]->cur_buf + count];
	  *((u_long*)(dst_addr + dst_off)) = ICP_BMA_TEST;
	}
      /* initialize to new snap */
      icp.complete[board] = FALSE;
      icp.timeout[board] = FALSE;
      if (inparam->set_snap > 0)
	{
	  /* set snap command */
	  if (cr_set_acqmd (board, CR_SNAP))
	    {
	      icp_error (board, "cr_set_acqmd");
	      error = -EIO;
	      cr_set_fcnt (board, 1);
	      goto error;
	    }
	  icp.opmode[board] = ICP_OP_SNAP;
	}
      else
	icp.opmode[board] = ICP_OP_TRIGGER;
      /* set timeout and wait or continue */
      if (flags & O_NONBLOCK)
	{
	  /* set timer BUT don't wait on queue */
	  if (cr_get_tmode (board) == CR_DISABLE)
	    error = icp_add_timer (board, icp.expire[board] *
				   (inparam->set_snap + 1), CAMERA);
	  else
	    /* in trigger mode you have much longer time to start */
	    error = icp_add_timer (board, ICP_WAIT_TRIG, CAMERA);
	  goto error;
	}
      else
	{
	  /* set timer AND wait on queue */
	  if (cr_get_tmode (board) == CR_DISABLE)
	    error = icp_sleep (board, icp.expire[board] *
			       (inparam->set_snap + 1), CAMERA);
	  else
	    /* in trigger mode you have much longer time to start */
	    error = icp_sleep (board, ICP_WAIT_TRIG, CAMERA);
	}
      if (icp.timeout[board])
	{
	  icp.timeout[board] = FALSE;
	  icp_printf ("icpci[%c]: timeout snap cycle!\n", 'E');
	  error = -ENODATA;
	  goto error;
	}
      /* test for magic sequence */
      for (count = 0; count < inparam->set_snap; count++)
	{
	  src_addr = icp.mem[board]->v_buf_base;
	  src_off = icp.mem[board]->buffer[icp.mem[board]->cur_buf + count];
	  if (icp_readl (((u_long*)(src_addr + src_off))) == ICP_MEM_TEST)
	    icp_printf ("icpci[%c]: mem buffer %d not valid!\n", 'E', count);
	  dst_addr = icp.bma[board]->v_buf_base;
	  dst_off = icp.bma[board]->buffer[icp.bma[board]->cur_buf + count];
	  if (*((u_long*)(dst_addr + dst_off)) == ICP_BMA_TEST)
	    icp_printf ("icpci[%c]: bma buffer %d not valid!\n", 'E', count);
	}
      /* reset operation mode */
      icp.opmode[board] = ICP_OP_NO;
      break;
      
      /* set acquisition mode to normal acquisition */ 
    case GIOC_CR_SET_NORMAL:
      /* disable trigger mode */
      if (cr_set_tmode (board, CR_DISABLE))
	{
	  icp_error (board, "cr_set_tmode");
	  error = -EIO;
	  goto error;
	}
      if (cr_get_smstat (board) == CR_ILACE)
	{
	  /* set starting field to EVEN_FIELD */
	  if (cr_set_fldsel (board, CR_EVEN_FIELD))
	    {
	      icp_error (board, "cr_set_fldsel");
	      error = -EIO;
	      goto error;
	    }
	}
      else
	{
	  /* set starting field to NEXT_FIELD */
	  if (cr_set_fldsel (board, CR_NEXT_FIELD))
	    {
	      icp_error (board, "cr_set_fldsel");
	      error = -EIO;
	      goto error;
	    }
	}
      break;

    /* set acquisition mode to triggered acquisition */
    case GIOC_CR_SET_TRIGGER:
      /* acquisition module ready for trigger mode ?! */
      switch (icp.cap[board]->module)
	{
	case ICP_AMVS:
	  if (vs_get_trigen (board) == VS_DISABLE)
	    {
	      if (vs_get_trigcy (board) == VS_IN_TRIGCY)
		{
		  /* in trigger cycle */
		  error = -EBUSY;
		  goto error;
		}
	      else
		{
		  if (vs_set_trigen (board, VS_ENABLE))
		    {
		    icp_error (board, "vs_set_trigen");
		    error = -EIO;
		    goto error;
		    }
		}
	    }
	  break;
	case ICP_AMDG:
	  if (dg_get_trigen (board) == DG_DISABLE)
	    {
	      if (dg_get_trigcy (board) == DG_IN_TRIGCY)
		{
		  /* in trigger cycle */
		  error = -EBUSY;
		  goto error;
		}
	      else
		{
		  if (dg_set_trigen (board, VS_ENABLE))
		    {
		      icp_error (board, "dg_set_trigen");
		      error = -EIO;
		      goto error;
		    }
		}
	    }
	  break;
	case ICP_AMPV:
	  if (pv_get_trigen (board) == PV_DISABLE)
	    {
	    if (pv_get_trigcy (board) == PV_IN_TRIGCY)
	      {
		/* in trigger cycle */
		error = -EBUSY;
		goto error;
	      }
	    else
	      {
		if (pv_set_trigen (board, PV_ENABLE))
		  {
		    icp_error (board, "pv_set_trigen");
		    error = -EIO;
		    goto error;
		  }
	      }
	    }
	  break;
	default:
	  error = -ENXIO;
	  goto error;
	}
      /* set starting field to NEXT_FIELD */
      if (cr_set_fldsel (board, CR_NEXT_FIELD))
	{
	  icp_error (board, "cr_set_fldsel");
	  error = -EIO;
	  goto error;
	}
      /* enable trigger mode */
      if (cr_set_tmode (board, CR_ENABLE))
	{
	  icp_error (board, "cr_set_tmode");
	  error = -EIO;
	  goto error;
	}
      break;

    /* trigger per software */
    case GIOC_CR_SET_SOFTTRIG:
      switch (icp.cap[board]->module)
	{
	case ICP_AMVS:
	  /* test for trigger mode */
	  if (vs_get_trigen (board) != VS_ENABLE)
	    { 
	      error = -EINVAL;
	      goto error;
	    }
	  /* test for trigger source */
	  if (vs_get_trigsrc (board) != VS_TRIG_SOFT)
	    {
	      error = -EINVAL;
	      goto error;
	    }
	  /* trigger board */
	  if (vs_sw_trig (board))
	    {
	      icp_error (board, "vs_sw_trig");
	      error = -EIO;
	      goto error;
	    }
	  break;
	case ICP_AMDG:
	  /* test for trigger mode */
	  if (dg_get_trigen (board) != DG_ENABLE)
	    { 
	      error = -EINVAL;
	      goto error;
	    }
	  /* test for trigger source */
	  if (dg_get_trigsrc (board) != DG_TRIG_SOFT)
	    {
	      error = -EINVAL;
	      goto error;
	    }
	  /* trigger */
	if (dg_sw_trig (board))
	  { 
	    icp_error (board, "dg_sw_trig");
	    error = -EIO;
	    goto error;
	  }
	break;
      case ICP_AMPV:
	/* test for trigger mode */
	if (pv_get_trigen (board) != PV_ENABLE)
	  { 
	    error = -EINVAL;
	    goto error;
	  }
	/* test for trigger source */
	if (pv_get_trigsrc (board) != PV_TRIG_SOFT)
	  {
	    error = -EINVAL;
	    goto error;
	  }
	/* trigger board */
	if (pv_sw_trig (board))
	  { 
	    icp_error (board, "pv_sw_trig");
	    error = -EIO;
	    goto error;
	  }
	break;
	default:
	  error = -ENXIO;
	  goto error;
	}
      break;

      /* set horizontal/vertical decimation; (clipping/padding) */
    case GIOC_CR_SET_HVDEC:
      if (icp.pci[board]->revision < ICP_REV_B)
	{
	/* old boards */
	if (inparam->set_hvdec != 1)
	  {
	    error = -ENXIO;
	    goto error;
	  }
	}
      else
	{
	  /* new boards */
	  switch (inparam->set_hvdec)
	    {
	    case 1:
	      inparam->set_hvdec = CR_DEC_BY_1;
	      break;	  
	    case 2:
	      inparam->set_hvdec = CR_DEC_BY_2;
	      break;	  
	    case 4:
	      inparam->set_hvdec = CR_DEC_BY_4;
	      break;	  
	    case 8:
	      inparam->set_hvdec = CR_DEC_BY_8;
	      break;	  
	    default:
	      error = -EINVAL;
	      goto error;
	    }
	  /* only horizontal and vertical decimation
	     with same factor is supported */
	  if (cr_set_hdec (board, inparam->set_hvdec) < 0 ||
	      cr_set_vdec (board, inparam->set_hvdec) < 0)
	    {
	      icp_error (board, "cr_set_[h|v]dec");
	      error = -EIO;
	      goto error;
	    }
	  /* recalculate buffer size */
	  if ((error = icp_bma_init (board)))
	    goto error;
	}
      break;
      
      /* AM-VS */
      
      /* read lookup table (256 values) */
    case GIOC_VS_GET_LUT:
      vs_get_lutdata (board, &outparam->get_vs_lut);
      break;
      
      /* write lookup table (256 values) */
    case GIOC_VS_SET_LUT:
      if (vs_set_lutdata (board, &inparam->set_vs_lut))
	{
	  icp_error (board, "vs_set_lutdata");
	  error = -EIO;
	  goto error;
	}
      break;

    /* linearize lookup table */
    case GIOC_VS_SET_LUT_LIN:
      if (vs_lin_lut (board))
	{
	  icp_error (board, "vs_lin_lut");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read negative reference voltage */
    case GIOC_VS_GET_NREF:
      outparam->get_vs_nref = vs_get_nref (board); 
      break;

      /* write negative reference voltage */
    case GIOC_VS_SET_NREF:
      ICP_PRINTF("icpci[%d]:           neg-ref: %d.\n", board,
		 inparam->set_vs_nref);
      if (inparam->set_vs_nref > VS_NREF_MAX)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (vs_set_nref (board, inparam->set_vs_nref))
	{
	  icp_error (board, "vs_set_nref");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read positive reference voltage */
    case GIOC_VS_GET_PREF:
      outparam->get_vs_pref = vs_get_pref (board); 
      break;

      /* write positive reference voltage */
    case GIOC_VS_SET_PREF:
      ICP_PRINTF("icpci[%d]:           pos-ref: %d.\n", board,
		 inparam->set_vs_pref);
      if (inparam->set_vs_pref > VS_PREF_MAX)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (vs_set_pref (board, inparam->set_vs_pref))
	{
	  icp_error (board, "vs_set_pref");
	  error = -EIO;
	  goto error;
	}
      break;

    /* AM-DIG */

    /* read 16 bit input LUT page numer */
    case GIOC_DG_GET_LUTPG:
      outparam->get_dg_lutpg = dg_get_lutpg (board); 
      break;

      /* write 16 bit input LUT page number */
    case GIOC_DG_SET_LUTPG:
      if (dg_set_lutpg (board, inparam->set_dg_lutpg))
	{
	  icp_error (board, "dg_set_lutpg");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read 16 bit input lookup table */
    case GIOC_DG_GET_LUT:
      dg_get_lut16 (board, &outparam->get_dg_lut16);
      break;

      /* write 16 bit input lookup table */
    case GIOC_DG_SET_LUT:
      if (dg_set_lut16 (board, &inparam->set_dg_lut16))
	{
	  icp_error (board, "dg_set_lut16");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read 8 bit optional input LUT page numer */
    case GIOC_DG_GET_OLUTPG:
      outparam->get_dg_olutpg = dg_get_olutpg (board); 
      break;

      /* write 8 bit optional input LUT page numer */
    case GIOC_DG_SET_OLUTPG:
      if (dg_set_olutpg (board, inparam->set_dg_olutpg))
	{
	  icp_error (board, "dg_set_olutpg");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read 8 bit optional input lookup table */
    case GIOC_DG_GET_OLUT:
      dg_get_lut8 (board, &outparam->get_dg_lut8);
      break;

      /* write 8 bit optional input lookup table */
    case GIOC_DG_SET_OLUT:
      if (dg_set_lut8 (board, &inparam->set_dg_lut8))
	{
	  icp_error (board, "dg_set_lut8");
	  error = -EIO;
	  goto error;
	}
      break;

    /* linearize lookup table */
    case GIOC_DG_SET_LUT_LIN:
      if (dg_get_modver (board) != DG_MODVER_8_TTL &&
	  dg_get_modver (board) != DG_MODVER_8_RS422)
	if (dg_lin_lut16 (board, DG_MODE_LINEAR))
	  {
	    error = -EIO;
	    goto error;
	  }
      break;

    /* normalize output to 8 bit (from 8, 10, 12, 16 bit input) */
    case GIOC_DG_SET_NORM_TO8:
      if (dg_get_modver (board) != DG_MODVER_8_TTL &&
	  dg_get_modver (board) != DG_MODVER_8_RS422)
	if (dg_lin_lut16 (board, DG_MODE_NORM_TO8))
	  {
	    error = -EIO;
	    goto error;
	  }
      /* set pixelsize an recalculate buffersizes */
      if (cr_set_pixsz (board, CR_PIXSZ_8))
	{
	  icp_error (board, "cr_set_pixsz");
	  error = -EIO;
	  goto error;
	}
      if ((error = icp_mem_init (board)))
	goto error;
      if (icp.bma_on)
	if ((error = icp_bma_init (board)))
	  goto error;
      break;

    /* normalize output to 16 bit (from 8, 10, 12, 16 bit input) */
    case GIOC_DG_SET_NORM_TO16:
      if (dg_get_modver (board) != DG_MODVER_8_TTL &&
	  dg_get_modver (board) != DG_MODVER_8_RS422)
	{
	  if (dg_lin_lut16 (board, DG_MODE_NORM_TO16))
	    {
	      error = -EIO;
	      goto error;
	    }
	}
      else
	{
	  error = -EIO;
	  goto error;
	}
      /* set pixelsize an recalculate buffersizes */
      if (cr_set_pixsz (board, CR_PIXSZ_16))
	{
	  icp_error (board, "cr_set_pixsz");
	  error = -EIO;
	  goto error;
	}
      if ((error = icp_mem_init (board)))
	goto error;
      if (icp.bma_on)
	if ((error = icp_bma_init (board)))
	  goto error;
      break;

    case GIOC_DG_GET_EXTIME:
      exclk = dg_get_exclk (board);
      extime = dg_get_extime (board);
      switch (exclk)
	{
	case DG_C_7_k:
	  extime <<= 11;
	  break;
	case DG_C_28_k:
	  extime <<= 9;
	  break;
	case DG_C_112_k:
	  extime <<= 7;
	  break;
	case DG_C_447_5_k:
	  extime <<= 5;
	  break;
	case DG_C_1_79_M:
	  extime <<= 3;
	  break;
	case DG_C_7_16_M:
	  extime <<= 1;
	  break;
	default:
	  error = -EIO;
	  goto error;
	}
      extime *= ICP_DLY_BASE;
      outparam->get_dg_extime.tv_sec = 0;
      outparam->get_dg_extime.tv_nsec = extime;
      break;

    case GIOC_DG_SET_EXTIME:
      if (inparam->set_dg_extime.tv_sec ||
	  inparam->set_dg_extime.tv_nsec > ICP_DLY_MAX ||
	  inparam->set_dg_extime.tv_nsec < ICP_DLY_MIN)
	{
	  error = -EINVAL;
	  goto error;
	}
      extime = inparam->set_dg_extime.tv_nsec / ICP_DLY_BASE;
      if (extime > (1 << 9) * 4098)
	exclk = DG_C_7_k, extime >>= 11; 
      else
	if (extime > (1 << 7) * 4098)
	  exclk = DG_C_28_k, extime >>= 9; 
	else
	  if (extime > (1 << 5) * 4098)
	    exclk = DG_C_112_k, extime >>= 7; 
	  else
	    if (extime > (1 << 3) * 4098)
	      exclk = DG_C_447_5_k, extime >>= 5; 
	    else
	      if (extime > (1 << 1) * 4098)
		exclk = DG_C_1_79_M, extime >>= 3; 
	      else
		exclk = DG_C_7_16_M, extime >>= 1; 
      
      ICP_PRINTF("icpci[%d]: shutter-time (ns): %ld.\n", board,
		 inparam->set_dg_extime.tv_nsec);
      if (dg_set_exclk (board, exclk))
	{
	  icp_error (board, "dg_set_exclk");
	  error = -EIO;
	  goto error;
	}
      if (dg_set_extime (board, extime))
	{
	  icp_error (board, "dg_set_extime");
	  error = -EIO;
	  goto error;
	}
      break;

    /* PCVision */

    /* read lookup table (256 values) */
    case HIOC_PV_GET_LUT:
      pv_get_lutdata (board, &outparam->get_pv_lut);
      break;

      /* write lookup table (256 values) */
    case HIOC_PV_SET_LUT:
      if (pv_set_lutdata (board, &inparam->set_pv_lut))
	{
	  icp_error (board, "pv_set_lutdata");
	  error = -EIO;
	  goto error;
	}
      break;

    /* linearize lookup table */
    case HIOC_PV_SET_LUT_LIN:
      if (pv_lin_lut (board))
	{
	  icp_error (board, "pv_lin_lut");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read negative reference voltage */
    case HIOC_PV_GET_NREF:
      outparam->get_pv_nref = pv_get_nref (board); 
      break;

      /* write negative reference voltage */
    case HIOC_PV_SET_NREF:
      ICP_PRINTF("icpci[%d]:           neg-ref: %d.\n", board,
		 inparam->set_pv_nref);
      if (inparam->set_pv_nref > PV_MAX_NREF)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (pv_set_nref (board, inparam->set_pv_nref))
	{
	  icp_error (board, "pv_set_nref");
	  error = -EIO;
	  goto error;
	}
      break;

    /* read positive reference voltage */
    case HIOC_PV_GET_PREF:
      outparam->get_pv_pref = pv_get_pref (board); 
      break;

      /* write positive reference voltage */
    case HIOC_PV_SET_PREF:
      ICP_PRINTF("icpci[%d]:           pos-ref: %d.\n", board,
		 inparam->set_pv_pref);
      if (inparam->set_pv_pref > PV_MAX_PREF)
	{
	  error = -EINVAL;
	  goto error;
	}
      if (pv_set_pref (board, inparam->set_pv_pref))
	{
	  icp_error (board, "pv_set_pref");
	  error = -EIO;
	  goto error;
	}
      break;
      
    default:
      error = -EINVAL;
      goto error;
    }

  return OK;

error:
  ICP_PRINTF("icpci[%d]:            error: %d.\n", board, error);

  return error;
}

